//
//  MyRingBuffer.m
//  SoundGenThru
//
//  Created by Justin Bakse on Thu Feb 19 2004.
//  This is very nearly a direct port of the AudioRingBuffer C++ class from Apple's AudioThru example.
//
    #import <Cocoa/Cocoa.h>
  //  #import <types.h>
    
#import "MyRingBuffer.h"


@implementation MyRingBuffer

-init
{
    [super init];
    //nothing
    return self;
}

-(void)dealloc
{
    if (mBuffer)
        free(mBuffer);
}

-(void)AllocateBytes:(int)bytesPerFrame forFrames:(int) capacityFrames
{
    if (mBuffer)
        free(mBuffer);
    
    mBytesPerFrame = bytesPerFrame;
    mCapacityFrames = capacityFrames;
    mCapacityBytes = bytesPerFrame * capacityFrames;
    mBuffer = (Byte *)malloc(mCapacityBytes);
    [self Clear];
    
}

-(void)Clear
{
    	memset(mBuffer, 0, mCapacityBytes);
	mStartOffset = 0;
	mStartFrame = 0;
	mEndFrame = 0;
}

-(Boolean) storeData:(const Byte *) data
          Frames:(UInt32) frames
            Time:(long) startFrame
{
SInt64 endFrame;
SInt64 newStart;
    if (frames > mCapacityFrames) return false;
    
    
    //warning "we have a critical region here"
    // reading and writing could well be in separate threads
    
     endFrame = startFrame + frames;
      if (startFrame >= mEndFrame + mCapacityFrames)
        // writing more than one buffer ahead -- fine but that means that everything we have is now too far in the past
        [self Clear];
    
    if (mStartFrame == 0) {
        // empty buffer
        mStartOffset = 0;
        mStartFrame = startFrame;
        mEndFrame = endFrame;
        memcpy(mBuffer, data, frames * mBytesPerFrame);
    } else {
        UInt32 offset0, offset1, nBytes;
        if (endFrame > mEndFrame) {
            // advancing (as will be usual with sequential stores)
            
            if (startFrame > mEndFrame) {
                // we are skipping some samples, so zero the range we are skipping
                offset0 = [self FrameOffset:mEndFrame];
                offset1 = [self FrameOffset:startFrame];
                if (offset0 < offset1)
                    memset(mBuffer + offset0, 0, offset1 - offset0);
                else {
                    nBytes = mCapacityBytes - offset0;
                    memset(mBuffer + offset0, 0, nBytes);
                    memset(mBuffer, 0, offset1);
                }
            }
            mEndFrame = endFrame;
            
            // except for the case of not having wrapped yet, we will normally
            // have to advance the start
             newStart = mEndFrame - mCapacityFrames;
            if (newStart > mStartFrame) {
                mStartOffset = (mStartOffset + (newStart - mStartFrame) * mBytesPerFrame) % mCapacityBytes;
                mStartFrame = newStart;
            }
        }
        // now everything is lined up and we can just write the new data
        offset0 = [self FrameOffset:startFrame];
        offset1 = [self FrameOffset:endFrame];
        if (offset0 < offset1)
            memcpy(mBuffer + offset0, data, offset1 - offset0);
        else {
            nBytes = mCapacityBytes - offset0;
            memcpy(mBuffer + offset0, data, nBytes);
            memcpy(mBuffer, data + nBytes, offset1);
        }
    }
    //printf("Store - buffer times: %.0f - %.0f, writing %.0f - %.0f\n", double(mStartFrame), double(mEndFrame), double(startFrame), double(endFrame));
  //   printf("store %p, %d, %ld\n", data, (int)frames, startFrame);
    
    return YES;
}

-(double)fetchData:(Byte *) data
          Frames:(UInt32) frames
            Time:(long) startFrame
{
    SInt64 endFrame = startFrame + frames;
    UInt32 offset0, offset1;
    
    if (startFrame < mStartFrame || endFrame > mEndFrame) {
        //printf("error - buffer times: %.0f - %.0f, reading for %.0f - %.0f\n", double(mStartFrame), double(mEndFrame), double(startFrame), double(endFrame));
        if (startFrame < mStartFrame)
            return (double)(startFrame - mStartFrame);
        else
            return (double)(endFrame - mEndFrame);
    }
    
     offset0 = [self FrameOffset:startFrame];
     offset1 = [self FrameOffset:endFrame];
    
    if (offset0 < offset1)
        memcpy(data, mBuffer + offset0, offset1 - offset0);
    else {
        UInt32 nBytes = mCapacityBytes - offset0;
        memcpy(data, mBuffer + offset0, nBytes);
        memcpy(data + nBytes, mBuffer, offset1);
    }
    
 //   printf("fetch %p, %d, %ld\n", data, (int)frames, startFrame);
    return (double)((mEndFrame - frames) - startFrame);
  
}

-(UInt32)FrameOffset:(SInt64) frameNumber
{ 
    return (mStartOffset + (UInt32)(frameNumber - mStartFrame) * mBytesPerFrame) % mCapacityBytes;
}


@end
