/* MySoundPort */

#import <Cocoa/Cocoa.h>
#include <CoreAudio/AudioHardware.h>
#include <unistd.h>

#import "MyRingBuffer.h"
#import "MyFilter.h"


typedef struct{
    AudioDeviceID	          device;           // the default output device
    char                          name[255];  // name of device
    unsigned int                  bufferSize; // bufferSize
    unsigned int                  interleaved;
    AudioStreamBasicDescription	  format;	   // info about the default device (sampleRate, etc.)
    unsigned int                  framesPerBuffer;       // samples to fill each cycle
    float                         maxValue;
    unsigned int                  safetyOffset;
    Boolean		          initialized; 	  // successful init?
} MyAudioDevice;

enum{
    kOff,
    kStarting,
    kRunning,
    kStopRequested
};
#define kSecondsInRingBuffer 2.


@interface MySoundPort : NSObject
{
    @public

    
    MyAudioDevice mOutputDevice;
    MyAudioDevice mInputDevice;
    int           mOutputProcState;
    int           mInputProcState;
    
    Boolean       mRunning;  
    Float64       mSampleRate;
    Float64       mLastInputSampleCount;
    Float64       mIODeltaSampleCount;
    Float64       mActualThruLatency; 
    Float64       mInToOutSampleOffset; //how far behind the scheduled output to grab from buffer to be safe.
    
    MyRingBuffer  *mInputBuffer, *mInputBuffer2;
    float         *mTempBuffer, *mLeftInput, *mRightInput, *mLeftOutput, *mRightOutput;

    IBOutlet MyFilter  *mFilter;
      
}

-(void) initMyAudioDevice: (MyAudioDevice*) d;
-(void) setup;
-(void) getAudioDevice:(MyAudioDevice*)d
                  from:(AudioDeviceID)whatDevice
               isInput:(int)isInput;


-(void) printDeviceReport: (MyAudioDevice*) d;
-(void) start;
-(void)	computeThruOffset;

-(void) checkError:(OSStatus) err
        withString:(char[255]) s;

-(MyRingBuffer*) mInputBuffer;
-(MyRingBuffer*) mInputBuffer2;

-(float) maxInputValue;


@end

OSStatus OutputIOProc (AudioDeviceID          inDevice, 
                       const AudioTimeStamp*  inNow,
                       const AudioBufferList* inInputData,
                       const AudioTimeStamp*  inInputTime, 
                       AudioBufferList*       outOutputData, 
                       const AudioTimeStamp*  inOutputTime,
                       void*                  inClientData);

OSStatus InputIOProc (AudioDeviceID          inDevice, 
                       const AudioTimeStamp*  inNow,
                       const AudioBufferList* inInputData,
                       const AudioTimeStamp*  inInputTime, 
                       AudioBufferList*       outOutputData, 
                       const AudioTimeStamp*  inOutputTime,
                       void*                  inClientData);



typedef struct {
    @defs(MySoundPort);
} mysoundportdef;