//
//  MyRingBuffer.m
//  SoundGenThru
//
//  Created by Justin Bakse on Thu Feb 19 2004.
//  This code is based largely on Apple's Sound Thru example. 
//  Simplifications have been made, and support for non interlaced devices has been added. (hopefully)
//

#import "MySoundPort.h"
#include <unistd.h>

@implementation MySoundPort
- init
{
    [super init];
  
    mInputBuffer = [[MyRingBuffer alloc] init];
    mInputBuffer2 = [[MyRingBuffer alloc] init];
    
    [self initMyAudioDevice: &mOutputDevice];
    [self initMyAudioDevice: &mInputDevice];
    
    mRunning = NO;
      
    return self;
}

-(void) initMyAudioDevice: (MyAudioDevice*) d
{
        d->device = 0;        
        d->name[0] = '\0'; 
        d->bufferSize = 0; 
        d->framesPerBuffer = 0;       
        d->initialized = NO; 	  
}

- (void)dealloc
{
    if (mTempBuffer)
        free(mTempBuffer);
    
    if (mLeftInput)
        free(mLeftInput);
    if (mRightInput)
        free(mRightInput);
    if (mLeftOutput)
        free(mLeftOutput);
    if (mRightOutput)
        free(mRightOutput);

    
    [mInputBuffer release]; 
    [mInputBuffer2 release];  
}

-(void) setup
{
    //get input
    [self getAudioDevice: &mInputDevice from:kAudioHardwarePropertyDefaultInputDevice isInput:YES];
    fprintf(stderr, "\n\nInput Device:\n\n");
    [self printDeviceReport: &mInputDevice];
    
    //get output
    [self getAudioDevice: &mOutputDevice from:kAudioHardwarePropertyDefaultOutputDevice isInput:NO];
    fprintf(stderr, "\n\nOutput Device:\n\n");
    [self printDeviceReport: &mOutputDevice];
    
}


-(void) getAudioDevice:(MyAudioDevice*)d
                  from:(AudioDeviceID)whatDevice
               isInput:(int)isInput
{
    OSStatus err = kAudioHardwareNoError;
    UInt32  count;
      UInt32 dataSize;
    
   
   
    ////////////////////////////////////////
    // GET THE OUTPUT DEVICE
    ////////////////////////////////////////
   
    //DEVICE ITSELF
    count = sizeof(d->device);
    err = AudioHardwareGetProperty(whatDevice,  &count, (void *) &d->device);
    [self checkError: err withString: "kAudioHardwarePropertyDefault(IN|OUT)Device"];
    
    
    //DEVICE NAME
    count = 255;
    err = AudioDeviceGetProperty(d->device, 0, isInput, kAudioDevicePropertyDeviceName,  &count, (void *) d->name);
    [self checkError: err withString: "kAudioDevicePropertyDeviceName"];
   
  

     
    //DEVICE BUFFER SIZE 
    count = sizeof(UInt32);
    err = AudioDeviceGetProperty(d->device, 0, isInput, kAudioDevicePropertyBufferSize, &count, &d->bufferSize);
    [self checkError: err withString: "kAudioDevicePropertyBufferSize"];
     
    
    // FORMAT
    count = sizeof(d->format);
    err = AudioDeviceGetProperty(d->device, 0, isInput, kAudioDevicePropertyStreamFormat, &count, &d->format);
    [self checkError: err withString: "kAudioDevicePropertyStreamFormat"];  
 
     
     dataSize = sizeof( d->maxValue );
    err = AudioDeviceGetProperty( d->device, 1, isInput, kAudioDevicePropertyVolumeDecibels , &dataSize, &d->maxValue );
    d->maxValue = pow(10, d->maxValue / 20.0);
    [self checkError: err withString: "kAudioDevicePropertyVolumeDecibels"]; 
    
   
   

    
    //DEVICE SAFETY OFFSET
    count = sizeof(d->safetyOffset);
    err = AudioDeviceGetProperty(d->device, 0, isInput, kAudioDevicePropertySafetyOffset, &count, &d->safetyOffset);
    [self checkError: err withString: "kAudioDevicePropertySafetyOffset"];
    
    //check if its interleaved
    d->interleaved = !(d->format.mFormatFlags & kAudioFormatFlagIsNonInterleaved);
    
    //Calc Samples Per Buffer
    d->framesPerBuffer = d->bufferSize / d->format.mBytesPerFrame;
    
    //VERIFY LEGAL SOUND FORMAT
    if (mOutputDevice.format.mFormatID != kAudioFormatLinearPCM) {
        fprintf(stderr, "ERROR: Device format not LinearPCM\n");
        return;
    }
    
    if (!(mOutputDevice.format.mFormatFlags & kLinearPCMFormatFlagIsFloat)) {
        fprintf(stderr, "ERROR: Device format not float.\n");
        return;
    }
    
    d->initialized = YES;
} 

-(void) printDeviceReport: (MyAudioDevice*) d
{
   
    fprintf(stderr, "Device Name(ID): %s (%d)\n", d->name, (int)d->device);
    fprintf(stderr, "initialized: %d\n", d->initialized);
    fprintf(stderr, "\n");
    fprintf(stderr, "DeviceFormat:\n");
    fprintf(stderr, "LinearPCM = %d\n", d->format.mFormatID == kAudioFormatLinearPCM);
    fprintf(stderr, "Float = %ld\n", d->format.mFormatFlags & kLinearPCMFormatFlagIsFloat);
    fprintf(stderr, "Interleaved = %d\n", d->interleaved);
    fprintf(stderr, "\n");
    fprintf(stderr, "mSampleRate = %g\n", d->format.mSampleRate);
    fprintf(stderr, "mFormatFlags = %08lX\n", d->format.mFormatFlags);
    fprintf(stderr, "mBytesPerPacket = %ld\n", d->format.mBytesPerPacket);
    fprintf(stderr, "mFramesPerPacket = %ld\n", d->format.mFramesPerPacket);
    fprintf(stderr, "mChannelsPerFrame = %ld\n", d->format.mChannelsPerFrame);
    fprintf(stderr, "mBytesPerFrame = %ld\n", d->format.mBytesPerFrame);
    fprintf(stderr, "mBitsPerChannel = %ld\n", d->format.mBitsPerChannel);
    fprintf(stderr, "\n");
    fprintf(stderr, "bufferSize = %d\n", d->bufferSize);
    fprintf(stderr, "safetyOffset = %d\n", d->safetyOffset);
    fprintf(stderr, "framesPerBuffer = %d\n", d->framesPerBuffer);
    fprintf(stderr, "\n\n");
    
}

-(void) start
{
    OSStatus err = kAudioHardwareNoError;
    
    if (mRunning) return;
    
    //Good Devices?
    if (mOutputDevice.device == kAudioDeviceUnknown) return;
    if (mInputDevice.device == kAudioDeviceUnknown) return;
   
    //Matched Devices?
    if (mInputDevice.format.mSampleRate != mOutputDevice.format.mSampleRate) {
        fprintf(stderr, "Error - sample rate mismatch: %f / %f\n", mInputDevice.format.mSampleRate, mOutputDevice.format.mSampleRate);
        return;
    }
    
    if (mInputDevice.format.mChannelsPerFrame != mOutputDevice.format.mChannelsPerFrame
	|| mInputDevice.format.mBytesPerFrame != mOutputDevice.format.mBytesPerFrame) {
        fprintf(stderr, "Error - format mismatch: %ld / %ld channels, %ld / %ld bytes per frame\n",
                mInputDevice.format.mChannelsPerFrame, mOutputDevice.format.mChannelsPerFrame,
                mInputDevice.format.mBytesPerFrame, mOutputDevice.format.mBytesPerFrame);
        fprintf(stderr, "It is really much better if you use the same device for input and output, or devices that have a similar format.\n");
        fprintf(stderr, "This program will try to get things to work out, but it is known that with some devices things are picky.\n");
        fprintf(stderr, "For instance going between Jack and another device will probably work at first, but break if you try to change the conncections in jack pilot.\n");
        fprintf(stderr, "Basicly, I'll give it a shot, but no one is promising anything.\n\n");
       // return;
    }
    
    if (mInputDevice.interleaved){
        [mInputBuffer AllocateBytes:mInputDevice.format.mBytesPerFrame forFrames: (int)(kSecondsInRingBuffer * mInputDevice.format.mSampleRate)];
        
    }else{
        
        [mInputBuffer AllocateBytes:mInputDevice.format.mBytesPerFrame forFrames: (int)(kSecondsInRingBuffer * mInputDevice.format.mSampleRate)];
        [mInputBuffer2 AllocateBytes:mInputDevice.format.mBytesPerFrame forFrames: (int)(kSecondsInRingBuffer * mInputDevice.format.mSampleRate)];
    }
    
    if (mOutputDevice.interleaved){
        mLeftInput = (float *)malloc((mOutputDevice.format.mBytesPerFrame/2) * mOutputDevice.framesPerBuffer);
        mRightInput = (float *)malloc((mOutputDevice.format.mBytesPerFrame/2) * mOutputDevice.framesPerBuffer);
        mLeftOutput = (float *)malloc((mOutputDevice.format.mBytesPerFrame/2) * mOutputDevice.framesPerBuffer);
        mRightOutput = (float *)malloc((mOutputDevice.format.mBytesPerFrame/2) * mOutputDevice.framesPerBuffer);
      
        memset(mLeftOutput, 0, (mOutputDevice.format.mBytesPerFrame/2) * mOutputDevice.framesPerBuffer);
        memset(mRightOutput, 0, (mOutputDevice.format.mBytesPerFrame/2) * mOutputDevice.framesPerBuffer);
    }else{
        mLeftInput = (float *)malloc(mOutputDevice.format.mBytesPerFrame * mOutputDevice.framesPerBuffer);
        mRightInput = (float *)malloc(mOutputDevice.format.mBytesPerFrame * mOutputDevice.framesPerBuffer);
        mLeftOutput = (float *)malloc(mOutputDevice.format.mBytesPerFrame * mOutputDevice.framesPerBuffer);
        mRightOutput = (float *)malloc(mOutputDevice.format.mBytesPerFrame * mOutputDevice.framesPerBuffer);
     
        memset(mLeftOutput, 0, (mOutputDevice.format.mBytesPerFrame) * mOutputDevice.framesPerBuffer);
        memset(mRightOutput, 0, (mOutputDevice.format.mBytesPerFrame) * mOutputDevice.framesPerBuffer);
    }
 

    //If we have interleaved input, we are going to need a temp buffer to hold enough to feed OUTPUT before we split it
    if (mInputDevice.interleaved){
        mTempBuffer = (float *)malloc(mOutputDevice.format.mBytesPerFrame * mOutputDevice.framesPerBuffer * 2); //we need to grab enough to feed our output (frames * bytes * 2 channels (its holding interleaved data))
    }
        
    mSampleRate = mInputDevice.format.mSampleRate;
    
    mRunning = YES;
    
    mInputProcState = kStarting;
    mOutputProcState = kStarting;
    
    err = AudioDeviceAddIOProc(mInputDevice.device, InputIOProc, (mysoundportdef *)self);
    [self checkError: err withString: "AudioDeviceAddIOProc-input"];
    
    err = AudioDeviceStart(mInputDevice.device, InputIOProc);
    [self checkError: err withString: "AudioDeviceStart-input"];
    
    
    err = AudioDeviceAddIOProc(mOutputDevice.device, OutputIOProc, (mysoundportdef *)self);
    [self checkError: err withString: "AudioDeviceAddIOProc-output"];
    
    err = AudioDeviceStart(mOutputDevice.device, OutputIOProc);
    [self checkError: err withString: "AudioDeviceStart-output"];
    
    do{
        usleep(1000);
    }while (mOutputProcState != kRunning);
    
    
    [self computeThruOffset];
}

-(void)	computeThruOffset
{
AudioTimeStamp inputTime, outputTime;

    OSStatus err = kAudioHardwareNoError;
    
    if (!mRunning) {
        mActualThruLatency = 0;
        mInToOutSampleOffset = 0;
        return;
    }
    
        
    err = AudioDeviceGetCurrentTime(mInputDevice.device, &inputTime);
    [self checkError: err withString: "AudioDeviceGetCurrentTime-inputTime"];
    err = AudioDeviceGetCurrentTime(mOutputDevice.device, &outputTime);
    [self checkError: err withString: "AudioDeviceGetCurrentTime-outputTime"];
    
    //fprintf(stderr, "in host: %20.0f  samples: %20.f  safety: %7d  buffer: %4d\n", (Float64)(inputTime.mHostTime), inputTime.mSampleTime, mInputDevice.safetyOffset, mInputDevice.framesPerBuffer);
    //fprintf(stderr, "out host: %20.0f  samples: %20.f  safety: %7d  buffer: %4d\n", (Float64)(outputTime.mHostTime), outputTime.mSampleTime, mOutputDevice.safetyOffset, mOutputDevice.framesPerBuffer);
   
    mActualThruLatency = (SInt32)(mInputDevice.safetyOffset + mInputDevice.framesPerBuffer +
                                mOutputDevice.safetyOffset + mOutputDevice.framesPerBuffer);
    
    mInToOutSampleOffset = mActualThruLatency + mIODeltaSampleCount;
    fprintf(stderr, "LATENCY REPORT:\n");
    fprintf(stderr, "thru latency: %.0f frames\ninToOutOffset: %0.f frames\n", mActualThruLatency, mInToOutSampleOffset);
    fprintf(stderr, "latency in seconds: %f\n\n", ((float)mActualThruLatency/(float)mOutputDevice.format.mChannelsPerFrame)/(float)mOutputDevice.format.mSampleRate);
}



-(void) checkError: (OSStatus) err withString:(char[255])  s
{
    char errCString[5];
    if (err != kAudioHardwareNoError) {
        strncpy ((char *)&errCString,(char *)&err, (size_t)4); errCString[4] = '\0';
        fprintf(stderr, "%s error: %d (%s)\n", s, (int) err, errCString);
        return;
    }
}

-(MyRingBuffer*) mInputBuffer
{
    return mInputBuffer;
}

-(MyRingBuffer*) mInputBuffer2
{
    return mInputBuffer2;
}

-(float) maxInputValue
{
    return mInputDevice.maxValue;
}

@end

OSStatus InputIOProc (AudioDeviceID          inDevice, 
                       const AudioTimeStamp*  inNow, 
                       const AudioBufferList* inInputData,
                       const AudioTimeStamp*  inInputTime, 
                       AudioBufferList*       outOutputData,
                       const AudioTimeStamp*  inOutputTime,
                       void*                  inClientData)

{
    MySoundPort *This = (MySoundPort *)inClientData;


    switch (This->mInputProcState) {
	case kStarting:
            This->mInputProcState = kRunning;
            break;
	case kStopRequested:
            AudioDeviceStop(inDevice, InputIOProc);
            This->mInputProcState = kOff;
            return noErr;
	default:
            break;
    }
    
    This->mLastInputSampleCount = inInputTime->mSampleTime;
    
    if (This->mInputDevice.interleaved){
        [[This mInputBuffer] storeData:(Byte *)inInputData->mBuffers[0].mData
                                Frames:This->mInputDevice.framesPerBuffer
                                  Time:(long)(inInputTime->mSampleTime)];
    }else{
       
        [[This mInputBuffer] storeData:(Byte *)inInputData->mBuffers[0].mData
                                Frames:This->mInputDevice.framesPerBuffer
                                  Time:(long)(inInputTime->mSampleTime)];
        
        [[This mInputBuffer2] storeData:(Byte *)inInputData->mBuffers[1].mData
                                Frames:This->mInputDevice.framesPerBuffer
                                  Time:(long)(inInputTime->mSampleTime)];        
    }
   
   
    return noErr;
    
}


OSStatus OutputIOProc (AudioDeviceID          inDevice, 
                       const AudioTimeStamp*  inNow, 
                       const AudioBufferList* inInputData,
                       const AudioTimeStamp*  inInputTime, 
                       AudioBufferList*       outOutputData,
                       const AudioTimeStamp*  inOutputTime,
                       void*                  inClientData)
{
    double delta;
     int n;
    
    MySoundPort *This = (MySoundPort *)inClientData;
 
    switch (This->mOutputProcState) {
        case kStarting:
             if (This->mInputProcState == kRunning) {
                 This->mOutputProcState = kRunning;
                 This->mIODeltaSampleCount = inOutputTime->mSampleTime - This->mLastInputSampleCount;
             }
            return noErr;
	case kStopRequested:
            AudioDeviceStop(inDevice, OutputIOProc);
            This->mOutputProcState = kOff;
            return noErr;
	default:
            break;
    }
    
    //FETCH from buffer into output
    //double delta = [[This mInputBuffer] fetchData:(Byte *)outOutputData->mBuffers[0].mData
    //                                       Frames:This->mOutputDevice.framesPerBuffer
    //                                         Time:(long)(inOutputTime->mSampleTime - This->mInToOutSampleOffset)];

 
    
    //Get input from Ring buffer and fit it into left and right input buffers
   
 

    if (This->mInputDevice.interleaved){
         delta = [[This mInputBuffer] fetchData:(Byte *)(This->mTempBuffer)
                                           Frames:This->mOutputDevice.framesPerBuffer
                                             Time:(long)(inOutputTime->mSampleTime - This->mInToOutSampleOffset)];
        
       
        for (n = 0; n < This->mOutputDevice.framesPerBuffer; n++){
            This->mLeftInput[n]  = This->mTempBuffer[n * 2 + 0];
            This->mRightInput[n] = This->mTempBuffer[n * 2 + 1];
        }
       
    }else{
            
         delta = [[This mInputBuffer] fetchData:(Byte *)(This->mLeftInput)
                                         Frames:This->mOutputDevice.framesPerBuffer
                                           Time:(long)(inOutputTime->mSampleTime - This->mInToOutSampleOffset)];
        
       
         delta = [[This mInputBuffer2] fetchData:(Byte *)(This->mRightInput)
                                          Frames:This->mOutputDevice.framesPerBuffer
                                            Time:(long)(inOutputTime->mSampleTime - This->mInToOutSampleOffset)];
        
    }
   
     //filter the inputs
   
    
    
    [This->mFilter filterInput:This->mLeftInput
                        Output:This->mLeftOutput
                   SampleCount:(UInt32)This->mOutputDevice.framesPerBuffer
                       Channel:0];
    
    [This->mFilter filterInput:This->mRightInput
                        Output:This->mRightOutput
                   SampleCount:(UInt32)This->mOutputDevice.framesPerBuffer
                       Channel:1];
   
    /*
    //display inputs and outputs
    [This->mInputView filterInput:This->mLeftInput
                               Output:This->mLeftOutput
                          SampleCount:(UInt32)This->mOutputDevice.framesPerBuffer
                              Channel:0];
     */
    
    //copy the resuts into the output (note this could be sped up in the noninterleaved case, by having the fitlers directly dump into the device)
  
    if (This->mOutputDevice.interleaved){
        int n;
        for (n = 0; n < This->mOutputDevice.framesPerBuffer; n++){
            ((float *)outOutputData->mBuffers[0].mData)[n * 2 + 0] = This->mLeftOutput[n];
            ((float *)outOutputData->mBuffers[0].mData)[n * 2 + 1] = This->mRightOutput[n];
        }
    
    }else{
        int n;
        for (n = 0; n < This->mOutputDevice.framesPerBuffer; n++){
            ((float *)outOutputData->mBuffers[0].mData)[n] = This->mLeftOutput[n];
            ((float *)outOutputData->mBuffers[1].mData)[n] = This->mRightOutput[n];
        }
    }

       return noErr;
}
